# -*- coding:utf-8 -*-

# Copyright (c) 2009-2013 - Simon Conseil

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import absolute_import

import codecs
import copy
import logging
import os
import sys

from clint.textui import colored
from distutils.dir_util import copy_tree
from jinja2 import Environment, FileSystemLoader, ChoiceLoader, PrefixLoader
from jinja2.exceptions import TemplateNotFound

from .image import Image
from .settings import get_thumb

INDEX_PAGE = "index.html"
SIGAL_LINK = "https://github.com/saimn/sigal"
THEMES_PATH = os.path.normpath(os.path.join(
    os.path.abspath(os.path.dirname(__file__)), 'themes'))


class Writer(object):
    """Generate html pages for each directory of images."""

    def __init__(self, settings, output_dir, theme=None):
        self.settings = settings
        self.output_dir = os.path.abspath(output_dir)
        self.theme = theme or settings['theme']
        self.logger = logging.getLogger(__name__)

        # optionally add index.html to the URLs
        self.url_ext = INDEX_PAGE if settings['index_in_url'] else ''

        # search the theme in sigal/theme if the given one does not exists
        if not os.path.exists(self.theme):
            self.theme = os.path.join(THEMES_PATH, self.theme)
            if not os.path.exists(self.theme):
                raise Exception("Impossible to find the theme %s" % self.theme)

        self.logger.info("Theme path : %s", self.theme)
        theme_relpath = os.path.join(self.theme, 'templates')
        default_loader = FileSystemLoader(
            os.path.join(THEMES_PATH, 'default', 'templates'))

        env = Environment(
            loader=ChoiceLoader([
                FileSystemLoader(theme_relpath),
                default_loader,  # implicit inheritance
                PrefixLoader({'!default': default_loader})  # explicit one
            ]),
            trim_blocks=True,
        )

        try:
            self.template = env.get_template(INDEX_PAGE)
        except TemplateNotFound:
            self.logger.error(colored.red('ERROR:') +
                              ' The index.html template was not found.')
            sys.exit(1)

        self.copy_assets()

        self.ctx = {
            'sigal_link': SIGAL_LINK,
            'theme': {'name': os.path.basename(self.theme)},
            'images': [],
            'albums': [],
            'breadcumb': ''
        }

    def copy_assets(self):
        """Copy the theme files in the output dir."""

        self.theme_path = os.path.join(self.output_dir, 'static')
        copy_tree(os.path.join(self.theme, 'static'), self.theme_path)

    def get_breadcumb(self, paths, relpath):
        """Paths to upper directories (with titles and links)."""

        tmp_path = relpath
        breadcumb = [((self.url_ext or '.'), paths[tmp_path]['title'])]

        while True:
            tmp_path = os.path.normpath(os.path.join(tmp_path, '..'))
            if tmp_path == '.':
                break

            url = os.path.relpath(tmp_path, relpath) + '/' + self.url_ext
            breadcumb.append((url, paths[tmp_path]['title']))

        return reversed(breadcumb)

    def write(self, paths, relpath):
        """Render the html page."""

        path = os.path.join(self.output_dir, relpath)
        index_url = os.path.relpath(self.output_dir, path) + '/' + self.url_ext
        self.logger.info("Output path : %s", path)

        ctx = copy.deepcopy(self.ctx)
        ctx.update({
            'settings': self.settings,
            'index_url': index_url,
            'index_title': paths['.']['title']
        })
        ctx['theme']['url'] = os.path.relpath(self.theme_path, path)

        if relpath != '.':
            ctx['breadcumb'] = self.get_breadcumb(paths, relpath)

        for i in paths[relpath]['img']:
            ctx['images'].append({'file': i,
                                  'thumb': get_thumb(self.settings, i)})

        for d in paths[relpath]['subdir']:
            dpath = os.path.normpath(os.path.join(relpath, d))
            alb_thumb = paths[dpath]['thumbnail']
            thumb_name = get_thumb(self.settings, alb_thumb)
            thumb_path = os.path.join(self.output_dir, dpath, thumb_name)
            self.logger.debug("Thumbnail path : %s", thumb_path)

            # generate the thumbnail if it is missing (if
            # settings['make_thumbs'] is False)
            if not os.path.exists(thumb_path):
                img = Image(os.path.join(self.output_dir, dpath, alb_thumb))
                img.thumbnail(thumb_path, self.settings['thumb_size'],
                              fit=self.settings['thumb_fit'],
                              quality=self.settings['jpg_quality'])

            ctx['albums'].append({
                'url': d + '/' + self.url_ext,
                'title': paths[dpath]['title'],
                'thumb': os.path.join(d, thumb_name)
            })

        # generate html page and save
        page = self.template.render(paths[relpath], **ctx)

        with codecs.open(os.path.join(path, INDEX_PAGE), 'w', 'utf-8') as f:
            f.write(page)



# -*- coding:utf-8 -*-

# Copyright (c) 2009-2013 - Simon Conseil

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import logging
import os

_DEFAULT_CONFIG = {
    'img_size': (640, 480),
    'make_thumbs': True,
    'thumb_prefix': '',
    'thumb_suffix': '',
    'thumb_size': (200, 150),
    'thumb_dir': 'thumbnails',
    'thumb_fit': True,
    'keep_orig': False,
    'orig_dir': 'original',
    'jpg_options': {'quality': 85, 'optimize': True, 'progressive': True},
    'copyright': '',
    'ext_list': ['.jpg', '.jpeg', '.JPG', '.JPEG', '.png'],
    'theme': 'colorbox',
    'index_in_url': False,
    'links': '',
    'google_analytics': ''
}


def get_thumb(settings, filename):
    """Return the path to the thumb."""

    path, filen = os.path.split(filename)
    name, ext = os.path.splitext(filen)
    return os.path.join(path, settings['thumb_dir'], settings['thumb_prefix'] +
                        name + settings['thumb_suffix'] + ext)


def read_settings(filename=None):
    """Read settings from a config file in the source_dir root."""

    logger = logging.getLogger(__name__)

    settings = _DEFAULT_CONFIG.copy()
    if filename:
        tempdict = {}
        execfile(filename, tempdict)
        settings.update((k, v) for k, v in tempdict.iteritems()
                        if k not in ['__builtins__'])

    for key in ('img_size', 'thumb_size'):
        size = settings[key]
        if size[1] > size[0]:
            size[0], size[1] = size[1], size[0]
            settings[key] = size
            logger.warning("The %s setting should be specified with the "
                           "largest value first.", key)

    return settings



# -*- coding: utf-8 -*-

thumb_suffix = '.tn'
thumb_size = (200, 150)
keep_orig = True

links = [('Example link', 'http://example.org'),
         ('Another link', 'http://example.org')]

copyright = "An example copyright message"



# -*- coding:utf-8 -*-

# Copyright (c) 2009-2013 - Simon Conseil

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import division

import logging
import os
import sys
from exceptions import IOError
from PIL import Image as PILImage
from PIL import ImageDraw, ImageOps, ImageFile

# EXIF specs Orientation constant
EXIF_ORIENTATION_TAG = 274


class Image(object):
    """Image container

    Prepare images: resize images, and create thumbnails with some options
    (squared thumbs, ...).

    :param filename: path to an image

    """

    def __init__(self, filename):
        self.filename = filename
        self.imgname = os.path.split(filename)[1]
        self.logger = logging.getLogger(__name__)

        with open(filename, 'rb') as fp:
            self.img = PILImage.open(fp)
            self.img.load()

        exif = self.img._getexif()
        if exif:
            # see: http://www.impulseadventure.com/photo/exif-orientation.html
            orientation = exif.get(EXIF_ORIENTATION_TAG)
            rotate_map = {3: 180, 6: -90, 8: 90}
            rotation = rotate_map.get(orientation)
            if rotation:
                self.img = self.img.rotate(rotation)

    def save(self, filename, **kwargs):
        """Save the image.

        Pass a dict with PIL options (quality, optimize, progressive). PIL can
        have problems saving large JPEGs if MAXBLOCK isn't big enough, So if
        we have a problem saving, we temporarily increase it. See
        http://github.com/jdriscoll/django-imagekit/issues/91

        """
        try:
            with quiet():
                self.img.save(filename, "JPEG", **kwargs)
        except IOError:
            old_maxblock = ImageFile.MAXBLOCK
            ImageFile.MAXBLOCK = self.img.size[0] * self.img.size[1]
            try:
                self.img.save(filename, "JPEG", **kwargs)
            finally:
                ImageFile.MAXBLOCK = old_maxblock

    def resize(self, size):
        """Resize the image.

        - check if the image format is portrait or landscape and adjust `size`.
        - compute the width and height ratio, and keep the min to resize the
          image inside the `size` box without distorting it.

        :param size: tuple with the (with, height) to resize

        """

        if self.img.size[0] > self.img.size[1]:
            newsize = size
        else:
            newsize = (size[1], size[0])

        wratio = newsize[0] / self.img.size[0]
        hratio = newsize[1] / self.img.size[1]
        ratio = min(wratio, hratio)
        newsize = (int(ratio * self.img.size[0]),
                   int(ratio * self.img.size[1]))

        if ratio < 1:
            self.img = self.img.resize(newsize, PILImage.ANTIALIAS)

    def add_copyright(self, text):
        "Add a copyright to the image"

        draw = ImageDraw.Draw(self.img)
        draw.text((5, self.img.size[1] - 15), '\xa9 ' + text)

    def thumbnail(self, filename, box, fit=True, quality=90):
        "Create a thumbnail image"

        if fit:
            self.img = ImageOps.fit(self.img, box, PILImage.ANTIALIAS)
        else:
            self.img.thumbnail(box, PILImage.ANTIALIAS)

        self.img.save(filename, quality=quality)


class quiet(object):
    """A context manager for suppressing the stderr activity of PIL's C
    libraries. Based on http://stackoverflow.com/a/978264/155370

    """
    def __enter__(self):
        self.stderr_fd = sys.__stderr__.fileno()
        self.null_fd = os.open(os.devnull, os.O_RDWR)
        self.old = os.dup(self.stderr_fd)
        os.dup2(self.null_fd, self.stderr_fd)

    def __exit__(self, *args, **kwargs):
        os.dup2(self.old, self.stderr_fd)
        os.close(self.null_fd)
        os.close(self.old)



# -*- coding:utf-8 -*-

# Copyright (c) 2009-2013 - Simon Conseil

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
sigal - simple static gallery generator
=======================================

sigal is yet another python script to prepare a static gallery of images:

* resize images, create thumbnails with some options (squared thumbs, ...).
* generate html pages.
"""

from __future__ import absolute_import

__author__ = "Simon Conseil"
__version__ = "0.3"
__license__ = "MIT"

import codecs
import logging
import os
import sys
from argh import ArghParser, arg
from logging import Formatter

from .gallery import Gallery
from .settings import read_settings

_DEFAULT_CONFIG_FILE = 'sigal.conf.py'


def init_logging(level=logging.INFO):
    """Logging config

    Set the level and create a more detailed formatter for debug mode.

    """
    logger = logging.getLogger(__name__)
    logger.setLevel(level)

    if level == logging.DEBUG:
        formatter = Formatter('%(levelname)s - %(message)s')
    else:
        formatter = Formatter('%(message)s')

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def init():
    """Copy a sample config file in the current directory."""

    from pkg_resources import resource_string
    conf = resource_string(__name__, 'templates/sigal.conf.py')

    with codecs.open('sigal.conf.py', 'w', 'utf-8') as f:
        f.write(conf)
    print "Sample config file created: sigal.conf.py"


@arg('input-dir', help='Input directory')
@arg('output-dir', nargs='?', default='_build',
     help='Output directory (default: _build/)')
@arg('-f', '--force', help="Force the reprocessing of existing images")
@arg('-v', '--verbose', help="Show all messages")
@arg('-d', '--debug', help="Show all message, including debug messages")
@arg('-c', '--config', help="Configuration file (default: sigal.conf.py in "
     "the current working directory)")
@arg('-t', '--theme', help="Specify a theme directory, or a theme name for "
     "the themes included with Sigal")
@arg('-n', '--ncpu', help="Number of cpu for parallel execution (default: 1)")
def build(input_dir, output_dir, debug=False, verbose=False, force=False,
          config=None, theme=None, ncpu=1):
    """Run sigal to process a directory. """

    level = (debug and logging.DEBUG) or (verbose and logging.INFO) \
            or logging.WARNING
    init_logging(level=level)
    logger = logging.getLogger(__name__)

    if not os.path.isdir(input_dir):
        logger.error("Input directory '%s' does not exist.", input_dir)
        sys.exit(1)

    if not os.path.relpath(output_dir, input_dir).startswith('..'):
        logger.error("Output directory should be outside of the input "
                     "directory.")
        sys.exit(1)

    logger.info("Reading settings ...")
    settings_file = config or _DEFAULT_CONFIG_FILE
    if not os.path.isfile(settings_file):
        logger.error("Settings file not found (%s)", settings_file)
        sys.exit(1)
    settings = read_settings(settings_file)

    gal = Gallery(settings, input_dir, output_dir, force=force, theme=theme,
                  ncpu=ncpu)
    gal.build()


@arg('path', nargs='?', default='_build',
     help='Directory to serve (default: _build/)')
def serve(path):
    """Run a simple web server."""

    import SimpleHTTPServer
    import SocketServer

    if os.path.exists(path):
        os.chdir(path)
        PORT = 8000
        Handler = SimpleHTTPServer.SimpleHTTPRequestHandler
        httpd = SocketServer.TCPServer(("", PORT), Handler, False)

        print " * Running on http://127.0.0.1:%i/" % PORT

        try:
            httpd.allow_reuse_address = True
            httpd.server_bind()
            httpd.server_activate()
            httpd.serve_forever()
        except KeyboardInterrupt:
            print '\nAll done!'
    else:
        sys.stderr.write("The '%s' directory doesn't exist.\n" % p)


def main():
    parser = ArghParser(description='Simple static gallery generator.',
                        version=__version__)
    parser.add_commands([init, build, serve])
    parser.dispatch()



# -*- coding:utf-8 -*-

from setuptools import setup, find_packages

requires = ['argh', 'clint', 'jinja2', 'Markdown']

try:
    from PIL import Image, ImageOps  # NOQA
except ImportError:
    requires += ['Pillow']

entry_points = {
    'console_scripts': ['sigal = sigal:main']
}

with open('README.rst') as f:
    README = f.read()

setup(
    name='sigal',
    version='0.3',
    url='https://github.com/saimn/sigal',
    license='MIT',
    author='Simon Conseil',
    author_email='contact@saimon.org',
    description='Simple static gallery generator',
    long_description=README,
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=requires,
    test_requires=['nose'],
    test_suite='nose.collector',
    entry_points=entry_points,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Multimedia :: Graphics :: Viewers',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



# -*- coding:utf-8 -*-

# Copyright (c) 2009-2013 - Simon Conseil

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import absolute_import

import codecs
import logging
import markdown
import os
import sys

from clint.textui import progress, colored
from multiprocessing import Pool
from os.path import join
from PIL import Image as PILImage

from .image import Image
from .settings import get_thumb
from .writer import Writer

DESCRIPTION_FILE = "index.md"


class PathsDb(object):
    """Container for all the information on the directory structure.

    All the info is stored in a dictionnary, `self.db`. This class also has
    methods to build this dictionnary.

    """

    def __init__(self, path, ext_list):
        self.basepath = path
        self.ext_list = ext_list
        self.logger = logging.getLogger(__name__)

    def get_subdirs(self, path):
        """Return the list of all sub-directories of path."""

        subdir = [os.path.normpath(os.path.join(path, sub))
                  for sub in self.db[path].get('subdir', [])]
        if subdir:
            return subdir + reduce(lambda x, y: x+y, map(self.get_subdirs, subdir))
        else:
            return []

    def build(self):
        "Build the list of directories with images"

        # The dict containing all information
        self.db = {
            'paths_list': [],
            'skipped_dir': []
        }

        # get information for each directory
        for path, dirnames, filenames in os.walk(self.basepath):
            relpath = os.path.relpath(path, self.basepath)

            # sort images and sub-albums by name
            filenames.sort(key=str.lower)
            dirnames.sort(key=str.lower)

            self.db['paths_list'].append(relpath)
            self.db[relpath] = {
                'img': [f for f in filenames
                        if os.path.splitext(f)[1] in self.ext_list],
                'subdir': dirnames
            }
            self.db[relpath].update(get_metadata(path))

        path_im = [path for path in self.db['paths_list']
                   if self.db[path]['img'] and path != '.']
        path_noim = [path for path in self.db['paths_list']
                     if not self.db[path]['img'] and path != '.']

        # dir with images: check the thumbnail, and find it if necessary
        for path in path_im:
            self.check_thumbnail(path)

        # dir without images, start with the deepest ones
        for path in reversed(sorted(path_noim, key=lambda x: x.count('/'))):
            for subdir in self.get_subdirs(path):
                # use the thumbnail of their sub-directories
                if self.db[subdir].get('thumbnail', ''):
                    self.db[path]['thumbnail'] = join(
                        os.path.relpath(subdir, path),
                        self.db[subdir]['thumbnail'])
                    break

            if not self.db[path].get('thumbnail', ''):
                # else remove all info about this directory
                self.logger.info("Directory '%s' is empty", path)
                self.db['skipped_dir'].append(path)
                self.db['paths_list'].remove(path)
                del self.db[path]
                parent = os.path.normpath(join(path, '..'))
                child = os.path.relpath(path, parent)
                self.db[parent]['subdir'].remove(child)

    def check_thumbnail(self, path):
        "Find the thumbnail image for a given path."

        # stop if it is already set and a valid file
        alb_thumb = self.db[path].setdefault('thumbnail', '')
        if alb_thumb and os.path.isfile(join(self.basepath, path, alb_thumb)):
            return

        # find and return the first landscape image
        for f in self.db[path]['img']:
            im = PILImage.open(join(self.basepath, path, f))
            if im.size[0] > im.size[1]:
                self.db[path]['thumbnail'] = f
                return

        # else simply return the 1st image
        if self.db[path]['img']:
            self.db[path]['thumbnail'] = self.db[path]['img'][0]
        else:
            self.db[path]['thumbnail'] = ''


class Gallery(object):
    "Prepare images"

    def __init__(self, settings, input_dir, output_dir, force=False,
                 theme=None, ncpu=1):
        self.settings = settings
        self.force = force
        self.ncpu = ncpu
        self.input_dir = os.path.abspath(input_dir)
        self.output_dir = os.path.abspath(output_dir)
        self.logger = logging.getLogger(__name__)
        self.writer = Writer(settings, output_dir, theme=theme)

        self.paths = PathsDb(self.input_dir, self.settings['ext_list'])
        self.paths.build()
        self.db = self.paths.db

    def build(self):
        "Create the image gallery"

        self.logger.info("Generate gallery in %s ...", self.output_dir)
        check_or_create_dir(self.output_dir)

        # Compute the label with for the progress bar. The max value is 48
        # character = 80 - 32 for the progress bar.
        label_width = max((len(p) for p in self.db['paths_list'])) + 1
        label_width = min(label_width, 48)

        # loop on directories in reversed order, to process subdirectories
        # before their parent
        for path in reversed(self.db['paths_list']):
            imglist = [join(self.input_dir, path, f)
                       for f in self.db[path]['img']]

            # output dir for the current path
            img_out = join(self.output_dir, path)
            check_or_create_dir(img_out)

            if len(imglist) != 0:
                self.process_dir(imglist, img_out, path,
                                 label_width=label_width)

            self.writer.write(self.db, path)

    def process_dir(self, imglist, outpath, dirname, label_width=20):
        """Process a list of images in a directory."""

        # Create thumbnails directory and optionally the one for original img
        check_or_create_dir(join(outpath, self.settings['thumb_dir']))

        if self.settings['keep_orig']:
            check_or_create_dir(join(outpath, self.settings['orig_dir']))

        # use progressbar if level is > INFO
        if self.logger.getEffectiveLevel() > 20:
            label = colored.green(dirname.ljust(label_width))
            img_iterator = progress.bar(imglist, label=label)
        else:
            img_iterator = iter(imglist)
            self.logger.info(":: Processing '%s' [%i images]",
                             colored.green(dirname), len(imglist))

        pool = Pool(processes=self.ncpu)

        try:
            # loop on images
            for f in img_iterator:
                filename = os.path.split(f)[1]
                outname = join(outpath, filename)
                self.logger.info(filename)

                if os.path.isfile(outname) and not self.force:
                    self.logger.info("%s exists - skipping", filename)
                else:
                    self.logger.info(filename)
                    pool.apply_async(worker_image,
                                     (f, outpath, self.settings)).get(9999)

            pool.close()
            pool.join()
        except KeyboardInterrupt:
            pool.terminate()
            sys.exit('Interrupted')


def worker_image(*args):
    try:
        return process_image(*args)
    except KeyboardInterrupt:
        return 'KeyboardException'


def process_image(filepath, outpath, settings):
    """Process one image: resize, create thumbnail."""

    filename = os.path.split(filepath)[1]
    outname = join(outpath, filename)

    img = Image(filepath)

    if settings['keep_orig']:
        img.save(join(outpath, settings['orig_dir'], filename),
                 **settings['jpg_options'])

    img.resize(settings['img_size'])

    if settings['copyright']:
        img.add_copyright(settings['copyright'])

    img.save(outname, **settings['jpg_options'])

    if settings['make_thumbs']:
        thumb_name = join(outpath, get_thumb(settings, filename))
        img.thumbnail(thumb_name, settings['thumb_size'],
                      fit=settings['thumb_fit'],
                      quality=settings['jpg_options']['quality'])


def get_metadata(path):
    """ Get album metadata from DESCRIPTION_FILE:

    - title
    - thumbnail image
    - description
    """

    descfile = join(path, DESCRIPTION_FILE)

    if not os.path.isfile(descfile):
        # default: get title from directory name
        meta = {
            'title': os.path.basename(path).replace('_', ' ')
            .replace('-', ' ').capitalize(),
            'description': '',
            'thumbnail': ''
        }
    else:
        md = markdown.Markdown(extensions=['meta'])

        with codecs.open(descfile, "r", "utf-8") as f:
            text = f.read()

        html = md.convert(text)

        meta = {
            'title': md.Meta.get('title', [''])[0],
            'description': html,
            'thumbnail': md.Meta.get('thumbnail', [''])[0]
        }

    return meta


def check_or_create_dir(path):
    "Create the directory if it does not exist"

    if not os.path.isdir(path):
        os.makedirs(path)
